import { LSPConnection } from './connection';
import { Signal } from '@phosphor/signaling';
import { PageConfig, URLExt } from '@jupyterlab/coreutils';
import { sleep, until_ready } from './utils';
/**
 * Each Widget with a document (whether file or a notebook) has its own DocumentConnectionManager
 * (see JupyterLabWidgetAdapter), keeping the virtual document spaces separate if a file is opened twice.
 */
export class DocumentConnectionManager {
    constructor() {
        this.connections = new Map();
        this.documents = new Map();
        this.ignored_languages = new Set();
        this.connected = new Signal(this);
        this.initialized = new Signal(this);
        this.disconnected = new Signal(this);
        this.closed = new Signal(this);
        this.documents_changed = new Signal(this);
    }
    connect_document_signals(virtual_document) {
        virtual_document.foreign_document_opened.connect((host, context) => {
            console.log('LSP: Connecting foreign document: ', context.foreign_document.id_path);
            this.connect_document_signals(context.foreign_document);
        });
        virtual_document.foreign_document_closed.connect((host, { foreign_document }) => {
            this.connections.get(foreign_document.id_path).close();
            this.connections.delete(foreign_document.id_path);
            this.documents.delete(foreign_document.id_path);
            this.documents_changed.emit(this.documents);
        });
        this.documents.set(virtual_document.id_path, virtual_document);
        this.documents_changed.emit(this.documents);
    }
    connect_socket(options) {
        let { virtual_document, language } = options;
        const uris = DocumentConnectionManager.solve_uris(virtual_document, language);
        let socket = new WebSocket(uris.socket);
        let connection = new LSPConnection({
            languageId: language,
            serverUri: uris.server,
            rootUri: uris.base,
            documentUri: uris.document,
            documentText: () => {
                // NOTE: Update is async now and this is not really used, as an alternative method
                // which is compatible with async is used.
                // This should be only used in the initialization step.
                // if (main_connection.isConnected) {
                //  console.warn('documentText is deprecated for use in JupyterLab LSP');
                // }
                return virtual_document.value;
            }
        }).connect(socket);
        connection.on('error', e => {
            console.warn(e);
            // TODO invalid now
            let error = e.length && e.length >= 1 ? e[0] : new Error();
            // TODO: those codes may be specific to my proxy client, need to investigate
            if (error.message.indexOf('code = 1005') !== -1) {
                console.warn('LSP: Connection failed for ' + virtual_document.id_path);
                console.log('LSP: disconnecting ' + virtual_document.id_path);
                this.closed.emit({ connection, virtual_document });
                this.ignored_languages.add(virtual_document.language);
                console.warn(`Cancelling further attempts to connect ${virtual_document.id_path} and other documents for this language (no support from the server)`);
            }
            else if (error.message.indexOf('code = 1006') !== -1) {
                console.warn('LSP: Connection closed by the server ' + virtual_document.id_path);
            }
            else {
                console.error('LSP: Connection error of ' + virtual_document.id_path + ':', e);
            }
        });
        this.connections.set(virtual_document.id_path, connection);
        return connection;
    }
    async retry_to_connect(options, reconnect_delay, retrials_left = -1) {
        let { virtual_document } = options;
        if (this.ignored_languages.has(virtual_document.language)) {
            return;
        }
        let interval = reconnect_delay * 1000;
        let success = false;
        while (retrials_left !== 0 && !success) {
            await this.connect(options)
                .then(() => {
                success = true;
            })
                .catch(e => {
                console.log(e);
            });
            console.log('LSP: will attempt to re-connect in ' + interval / 1000 + ' seconds');
            await sleep(interval);
            // gradually increase the time delay, up to 5 sec
            interval = interval < 5 * 1000 ? interval + 500 : interval;
        }
    }
    async connect(options) {
        let connection = this.connect_socket(options);
        connection.on('serverInitialized', capabilities => {
            this.initialized.emit({ connection, virtual_document });
        });
        let { virtual_document, document_path } = options;
        await until_ready(() => {
            // @ts-ignore
            return connection.isConnected;
        }, 50, 50).catch(() => {
            throw Error('LSP: Connect timed out for ' + virtual_document.id_path);
        });
        console.log('LSP:', document_path, virtual_document.id_path, 'connected.');
        connection.on('close', closed_manually => {
            if (!closed_manually) {
                console.warn('LSP: Connection unexpectedly disconnected');
                this.retry_to_connect(options, 0.5).catch(console.warn);
            }
            else {
                console.warn('LSP: Connection closed');
                this.closed.emit({ connection, virtual_document });
            }
        });
        this.connected.emit({ connection, virtual_document });
        return connection;
    }
    close_all() {
        for (let [id_path, connection] of this.connections.entries()) {
            let virtual_document = this.documents.get(id_path);
            connection.close();
            // TODO: close() should trigger the closed event, but it does not seem to work, hence manual trigger below:
            this.closed.emit({ connection, virtual_document });
        }
        this.connections.clear();
    }
}
(function (DocumentConnectionManager) {
    function solve_uris(virtual_document, language) {
        const wsBase = PageConfig.getBaseUrl().replace(/^http/, 'ws');
        const rootUri = PageConfig.getOption('rootUri');
        const virtualDocumentsUri = PageConfig.getOption('virtualDocumentsUri');
        const baseUri = virtual_document.has_lsp_supported_file
            ? rootUri
            : virtualDocumentsUri;
        return {
            base: baseUri,
            document: URLExt.join(baseUri, virtual_document.uri),
            server: URLExt.join('ws://jupyter-lsp', language),
            socket: URLExt.join(wsBase, 'lsp', language)
        };
    }
    DocumentConnectionManager.solve_uris = solve_uris;
})(DocumentConnectionManager || (DocumentConnectionManager = {}));
//# sourceMappingURL=connection_manager.js.map